# This file was auto-generated by Fern from our API Definition.

import typing

from .credit_card_credential_response import CreditCardCredentialResponse
from .password_credential_response import PasswordCredentialResponse

CredentialResponseCredential = typing.Union[PasswordCredentialResponse, CreditCardCredentialResponse]
